;; Mastermind
;; Author: Jon Ludwig <jcl1827@cs.rit.edu>
;;
;; Version: Info-minimax
;; This version solves mastermind with a modified version of Knuth's algorithm
;; Instead of computing the minimum number of removals for each possible guess
;; we compute the minimum amount of information given by each guess

(define sequences #f)
(define goal #f)
(define color-list #f)

;; solves an instace of mastermind
(define (mastermind pegs colors sequence)
  (set! goal sequence)
  (set! color-list (expand-list colors))
  (set! sequences (generate-sequences pegs color-list '()))
  ;; bounds checking
  (if (and (eq? (length sequence) pegs) (<= (apply max sequence) colors))
      (guess (initial-guess pegs) pegs sequences)
      (display "invalid arguments")
      )
  )

;; make a guess and continue until solved or failure
(define (guess g pegs possibles)
  (if (null? g) (display "failed!")
      (let ((rating (rate-sequence g goal))) ;; here is where we calculate the rating against the goal sequence (substitute a human here if you like)
        (display 'guess )
        (display g)
        (newline)
        (if (eq? (get-exact-matches rating) pegs) (display "solved!")
            (let ((new-possibles (prune rating possibles)))
              (guess (minmax new-possibles pegs) pegs new-possibles)
              )
            )
        )
      )
  )

;; calculates the maximum amount of information for each guess in guess-set
(define (minmax guess-set pegs)
  (car (max-guess (map (lambda (el) (cons el (list (min-info el guess-set pegs)))) guess-set)))
  )
  
;; returns the guess with the max amount of information
(define (max-guess guesses)
  (if (null? (cdr guesses)) (car guesses)
      (let ((other (max-guess (cdr guesses))))
        (if (> (cadr (car guesses)) (cadr other))
            (car guesses)
            other
            )
        )
      )
  )

;; calculates the minmum number of removals for given guess over all possible responses
(define (min-info guess guess-set pegs)
  (apply min (map (lambda (el) (/ (log (/ (length guess-set) (max 1 (length (prune (cons guess el) guess-set))))) (log 2))) (possible-responses pegs)))
  )


;; return a list of all the possible responses (black + white pegs) given a number of pegs
;; * this is a complete list and will include certain combinations which are impossible to achieve
(define (possible-responses pegs)
  (flatmap (lambda (first) (flatmap (lambda (second) (if (<= (+ first second) pegs) (list (cons first (list second))) '())) (cons 0 (expand-list pegs)))) (cons 0 (expand-list pegs)))
  )

;; prune sequences which cannot be solutions
;; * for a sequence to be a possible solution, the score against the given guess must be the same as that guess against the solution
(define (prune guess guess-set)
  (filter (lambda (el) (match-rating el guess)) guess-set)
  )

;; determine if a seq is equal in score to another sequence (white and black pegs score is the same)
(define (match-rating seq match-seq)
  ;; this is much quicker
  (let ((exact-matches (find-exact-matches seq (get-sequence match-seq))))
    (if (eq? exact-matches (get-exact-matches match-seq))
        (eq? (- (find-matches seq (get-sequence match-seq)) exact-matches) (get-color-matches match-seq))
        #f
        )
    )
  )

;; generate an initial guess
(define (initial-guess pegs)
  (append (repeat 1 (floor (/ pegs 2))) (repeat 2 (- pegs (floor (/ pegs 2)))))
  )

;; append a rating (white + black pegs) to a sequence matched against the given goal
(define (rate-sequence seq goal-seq)
  (let ((exact-matches (find-exact-matches seq goal-seq)))
    (let ((color-matches (- (find-matches seq goal-seq) exact-matches)))
     (build-state seq exact-matches color-matches)
      )
    )
  )

;; return the number of color matches between seq and goal-seq (white pegs + black pegs)
(define (find-matches seq goal-seq)
  (apply + (map (lambda (el) (min (number-of-occurances goal-seq el) (number-of-occurances seq el))) color-list))
  )
  
;; return the number of exact matches between seq and goal-seq (black pegs)
(define (find-exact-matches seq goal-seq)
  (if (null? seq) 0
      (+ (if (eq? (car seq) (car goal-seq)) 1 0) (find-exact-matches (cdr seq) (cdr goal-seq)))
      )
  )

;; return the number of occurances of a color in a sequence
(define (number-of-occurances seq color)
  (if (null? seq) 0
      (+ (if (eq? (car seq) color) 1 0) (number-of-occurances (cdr seq) color))
      )
  )

;; generate a list of all possible sequences (colors^pegs combinations)
(define (generate-sequences pegs colors sequence)
  (if (<= pegs 0) (list sequence)
      (flatmap (lambda (el) (generate-sequences (- pegs 1) colors (append sequence (list el)))) colors)
      )
  )

;; generate a list from 1 to the number, for easier manipulation
(define (expand-list num)
  (if (<= num 0) '()
      (cons num (expand-list (- num 1)))
      )
  )

;; utility functions
(define (filter fun lst)
  (if (null? lst) '()
      (append (if (fun (car lst))
                  (list (car lst))
                  '())
              (filter fun (cdr lst)))
      )
  )

(define (flatmap fun lst)
  (if (null? lst) '()
      (append (fun (car lst)) (flatmap fun (cdr lst)))
      )
  )

(define (repeat what how-many)
  (if (<= how-many 0) '()
      (cons what (repeat what (- how-many 1)))
      )
  )

;; state manipulation and accessor functions
(define (build-state sequence exact-matches color-matches)
  (append (list sequence) (cons exact-matches (list color-matches)))
  )

(define (get-sequence state)
  (car state)
  )

(define (get-exact-matches state)
  (cadr state)
  )

(define (get-color-matches state)
  (caddr state)
  )
