;; Mastermind
;; Author: Jon Ludwig <jcl1827@cs.rit.edu>
;;
;; Version: Knuth-simple
;; This version solves mastermind with a simplified version of Knuth's algorithm
;; * Instead of choosing the pruned guess with the minimum number of maximum

(define sequences #f)
(define goal #f)
(define color-list #f)

;; solves an instance of mastermind
(define (mastermind pegs colors sequence)
  (set! goal sequence)
  (set! color-list (generate-color-list colors))
  (set! sequences (generate-sequences pegs color-list '()))
  ;; bounds checking
  (if (and (eq? (length sequence) pegs) (<= (apply max sequence) colors))
      (guess (initial-guess pegs) pegs)
      (display "invalid arguments")
      )
  )

;; make a guess and continue until solved or failure
(define (guess g pegs)
  (if (null? g) (display "failed!")
      (let ((rating (rate-sequence g goal))) ;; here is where we calculate the rating against the goal sequence (substitute a human here if you like)
        (display 'guess )
        (display g)
        (newline)
        (if (eq? (get-exact-matches rating) pegs) (display "solved!")
            (guess (car (prune rating)) pegs)
            )
        )
      )
  )

;; prune sequences which cannot be solutions
;; * for a sequence to be a possible solution, the score against the given guess must be the same as that guess against the solution
;; * note: affects global "sequences"
(define (prune guess)
  (set! sequences (filter (lambda (el) (match-rating el guess)) sequences))
  (if (null? sequences) '(())
      sequences
      )
  )

;; determine if a seq is equal in score to another sequence (white and black pegs score is the same)
(define (match-rating seq match-seq)
  (let ((rating (rate-sequence seq (get-sequence match-seq))))
    (and (eq? (get-color-matches rating) (get-color-matches match-seq))
         (eq? (get-exact-matches rating) (get-exact-matches match-seq)))
    )
  )

;; generate an initial guess
(define (initial-guess pegs)
  (append (repeat 1 (floor (/ pegs 2))) (repeat 2 (- pegs (floor (/ pegs 2)))))
  )

;; append a rating (white + black pegs) to a sequence matched against the given goal
(define (rate-sequence seq goal-seq)
  (let ((exact-matches (find-exact-matches seq goal-seq)))
    (let ((color-matches (- (find-matches seq goal-seq) exact-matches)))
     (build-state seq exact-matches color-matches)
      )
    )
  )

;; return the number of color matches between seq and goal-seq (white pegs + black pegs)
(define (find-matches seq goal-seq)
  (apply + (map (lambda (el) (min (number-of-occurances goal-seq el) (number-of-occurances seq el))) color-list))
  )
  
;; return the number of exact matches between seq and goal-seq (black pegs)
(define (find-exact-matches seq goal-seq)
  (if (null? seq) 0
      (+ (if (eq? (car seq) (car goal-seq)) 1 0) (find-exact-matches (cdr seq) (cdr goal-seq)))
      )
  )

;; return the number of occurances of a color in a sequence
(define (number-of-occurances seq color)
  (if (null? seq) 0
      (+ (if (eq? (car seq) color) 1 0) (number-of-occurances (cdr seq) color))
      )
  )

;; generate a list of all possible sequences (colors^pegs combinations)
(define (generate-sequences pegs colors sequence)
  (if (<= pegs 0) (list sequence)
      (flatmap (lambda (el) (generate-sequences (- pegs 1) colors (append sequence (list el)))) colors)
      )
  )

;; generate a list from 1 to the number of colors, for easier manipulation
(define (generate-color-list colors)
  (if (<= colors 0) '()
      (cons colors (generate-color-list (- colors 1)))
      )
  )

;; utility functions
(define (filter fun lst)
  (if (null? lst) '()
      (append (if (fun (car lst))
          (list (car lst))
          '())
            (filter fun (cdr lst)))
      )
  )

(define (flatmap fun lst)
  (if (null? lst) '()
      (append (fun (car lst)) (flatmap fun (cdr lst)))
      )
  )

(define (repeat what how-many)
  (if (<= how-many 0) '()
      (cons what (repeat what (- how-many 1)))
      )
  )

;; state manipulation and accessor functions
(define (build-state sequence exact-matches color-matches)
  (append (list sequence) (cons exact-matches (list color-matches)))
  )

(define (get-sequence state)
  (car state)
  )

(define (get-exact-matches state)
  (cadr state)
  )

(define (get-color-matches state)
  (caddr state)
  )